/* Inter thread communication benchmark
 *
 * Copyright (C) 2006 Stefan Westerfeld
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "birnet/birnet.h"
#include <sys/poll.h>
#include <semaphore.h>

using namespace Birnet;

struct Buffer
{
  vector<unsigned char> mem;
  int rpos, wpos;

  enum {
    TAG_INTEGER
  };
  struct Entry {
    int tag;
    union
    {
	struct
	{
	    int value;
	} i;
    };
  };

  Buffer() :
    rpos (0),
    wpos (0)
  {
  }

  void
  put_int (int arg)
  {
    int new_wpos = wpos + sizeof (Entry);
    if (new_wpos > mem.size())
      mem.resize (new_wpos);
    Entry *entry = reinterpret_cast<Entry *> (&mem[wpos]);
    entry->tag = TAG_INTEGER;
    entry->i.value = arg;
    wpos = new_wpos;
  }

  int
  get_int ()
  {
    int new_rpos = rpos + sizeof (Entry);
    g_return_val_if_fail (new_rpos <= mem.size(), 0);
    Entry *entry = reinterpret_cast<Entry *> (&mem[rpos]);
    rpos = new_rpos;
    g_return_val_if_fail (entry->tag == TAG_INTEGER, 0);
    return entry->i.value;
  }

  void
  reset()
  {
    rpos = 0;
    wpos = 0;
  }
};

struct CondSync {
  Cond condition;
  Mutex mutex;
  guint n_wakeups;

  CondSync()
    : n_wakeups (0)
  {
  }

  ~CondSync()
  {
    g_assert (n_wakeups == 0);
  }

  void
  wait()
  {
    AutoLocker locker (mutex);
    while (n_wakeups == 0)
      condition.wait (mutex);
    n_wakeups--;
  }

  void
  wakeup()
  {
    AutoLocker locker (mutex);
    n_wakeups++;
    condition.signal();
  }
};

struct PipeSync {
  int pfds[2];
  PipeSync()
  {
    pipe (pfds);
  }

  void
  wait()
  {
    struct pollfd poll_fds[1];
    poll_fds[0].fd = pfds[0];
    poll_fds[0].events = POLLIN;
    poll_fds[0].revents = 0;
    for (;;)
      {
	if (poll (poll_fds, 1, -1) > 0)
	  {
	    char c;
	    while (read (pfds[0], &c, 1) != 1)
	      ;
	    return;
	  }
      }
  }

  void
  wakeup()
  {
    while (write (pfds[1], "W", 1) != 1)
      ;
  }
};

struct SemSync {
  sem_t semaphore;

  SemSync()
  {
    sem_init (&semaphore, /* do not share across processes */ 0, /* initial value */ 0);
  }

  void
  wait()
  {
    sem_wait (&semaphore);
  }

  void
  wakeup()
  {
    sem_post (&semaphore);
  }
};


struct Method
{
  /******* modify synchronization primitive here *******/
  PipeSync server;
  PipeSync client;

  Buffer arg_buffer;
  Buffer result_buffer;

  int
  call (int arg)
  {
    arg_buffer.reset();
    arg_buffer.put_int (arg);
    server.wakeup();
    client.wait();
    return result_buffer.get_int();
  }

  int
  call_impl (int arg)
  {
    return arg * 2;
  }

  bool
  server_iteration()
  {
    server.wait();
    bool keep_running = dispatch();
    client.wakeup();

    return keep_running;
  }

  bool
  dispatch()
  {
    int arg = arg_buffer.get_int();
    int result = call_impl (arg);

    result_buffer.reset();
    result_buffer.put_int (result);

    return (result > 0);
  }

} method;

class ClientThread : public Thread
{
public:
  ClientThread() :
    Thread ("ClientThread")
  {
  }
protected:
  void
  run()
  {
    for (int i = 0; i < 1000000; i++)
      {
	int result = method.call (42);
	g_assert (result == 84);
      }

    method.call (0);
  }
};

class ServerThread : public Thread
{
public:
  ServerThread() :
    Thread ("ServerThread")
  {
  }
protected:
  void
  run()
  {
    while (method.server_iteration())
      ;
  }
};

int
main (int    argc,
      char **argv)
{
  birnet_init (&argc, &argv, "itc");
  ClientThread *client_thread = new ClientThread();
  ServerThread *server_thread = new ServerThread();

  client_thread->start();
  server_thread->start();
  client_thread->wait_for_exit();
  server_thread->wait_for_exit();

  client_thread->unref();
  server_thread->unref();
}
